﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace NSEventDemo.NSEvent
{
    public sealed class NSEM
    {
        private static NSEM instance;
        public static NSEM Default()
        {
            if (instance == null)
                instance = new NSEM();
            return instance;
        }

        private readonly Dictionary<Type, List<object>> typeInstances;
        private NSEM()
        {
            this.typeInstances = new Dictionary<Type, List<object>>();
        }

        internal void Subscribe(object instance)
        {
            bool[] result = new bool[1];
            foreach (var irfc in Helper.GetNotificationReceivers(instance, result))
            {
                if (!typeInstances.ContainsKey(irfc))
                    typeInstances.Add(irfc, new List<object>());

                typeInstances[irfc].Add(instance);
            }

            if (!result[0])
                throw new InvalidOperationException();
        }

        public void Unsubscribe(object instance)
        {
            this.typeInstances[instance.GetType()].Remove(instance);
        }

        internal void NotifyAllInstances<S>(object sender, dynamic argument)
        {
            if (!typeInstances.ContainsKey(typeof(S)))
                return;
            var instances = typeInstances[typeof(S)];
            for (int i = 0; i < instances.Count; i++)
            {
                dynamic objectInstance = instances[i];
                objectInstance.Notify(sender, argument);
            }
        }

        public static dynamic New<T>(params dynamic?[]? arguments)
        {
            if (Helper.IsNotificationReceiver<T>())
            {
                dynamic instance = Activator.CreateInstance(typeof(T), arguments); ;
                NSEM.Default().Subscribe(instance);
                return instance;
            }

            object[] injectedArguments = new object[arguments.Length + 1];
            injectedArguments[0] = new BasicNotificationAdapter<T>();
            Array.Copy(arguments, 0, injectedArguments, 1, arguments.Length);

            return Activator.CreateInstance(typeof(T), injectedArguments);
        }
    }
}
